/*  This is the LCDproc driver for CrystalFontz 633 devices
    (get yours from http://crystalfontz.com)

    Copyright (C) 2002 David GLAUDE

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 */


/*
 * Driver status
 * 04/04/2002: Working driver
 * 05/06/2002: Reading of return value
 * 02/09/2002: KeyPad handeling and return string
 * 03/09/2002: New icon incorporated
 *
 * THINGS NOT DONE:
 * + No checking if right hardware is connected (firmware/hardware)
 * + No BigNum (but screen is too small ???)
 * + No support for multiple instance (require private structure)
 * + No cache of custom char usage (like in MtxOrb)
 *
 * THINGS DONE:
 * + Stopping the live reporting (of temperature)
 * + Stopping the reporting of temp and fan (is it necessary after reboot)
 * + Use of library for hbar and vbar (good but library could be better)
 * + Support for keypad (Using a KeyRing)
 *
 * THINGS TO DO:
 * + Make the caching at least for heartbeat icon
 * + Create and use the library (for custom char handeling)
 *
 * 10/4/04 added temp and fan functions,
 * featuring awkward, and perhaps widely held to be bad form
 * declarations of globals in CFontz633io.c, which are then assigned
 * values in CFontz633.c, and used in CFontz633io.c
 * The 4 *633*.? files in server/drivers have changed dependencies
 * that are as not yet reflected in the Makefile / automake files:
 * the above plus say ../sceen.h in io.c.
 * make clean; make from the top has been required.
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>


#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "lcd.h"
#include "CFontz633.h"
#include "CFontz633io.h"
#include "report.h"
#include "lcd_lib.h"
#include "timing.h"
#include "../screen.h"
#include "../widget.h"

#define CF633_KEY_UP       1
#define CF633_KEY_DOWN     2
#define CF633_KEY_LEFT     3
#define CF633_KEY_RIGHT    4
#define CF633_KEY_ENTER    5
#define CF633_KEY_ESCAPE   6

static int custom = 0;
typedef enum {
	hbar = 1,
	vbar = 2,
	cust = 3,
} custom_type;

static int fd;
static char *framebuf = NULL;
static char *old = NULL;
static int width = 0;
static int height = 0;
static int cellwidth = DEFAULT_CELL_WIDTH;
static int cellheight = DEFAULT_CELL_HEIGHT;
static int contrast = DEFAULT_CONTRAST;
static int brightness = DEFAULT_BRIGHTNESS;
static int offbrightness = DEFAULT_OFFBRIGHTNESS;
static int newfirmware = 0;

//#define TIME_INIT  	// define to build all init timing chunks  
/*
 * TIME_INIT / time_init 
 *
 * does the init time varry much with # of sensors? 
 * 	--> calculate a sleep(depends on # of sensors) ?
 * does the init time varry in any significant way when run at boot 
 * amid load through init.d?
 * 	which lcdproc client is slowing my system clock?
 * 	will this measuring do the same?
 *
 *	CFA_633.pdf p.13 : The host should wait for an acknowledge packet
 * before sending the next command packet.  The CFA_633 will respond to all packets
 * within 250mS, [ er... reboot=~2 seconds ] so the host software should sop
 * waiting and retry the packet if the CFA_633 fails to respond in 250mS.  The host 
 * software should report an error if the if a packet is not acknowledged after several
 * tries.  This could be indicative of a hardware problem.
 *
 */
#ifdef TIME_INIT
	int time_init=4;  // 1 = init once, zero reps, with timing report
			 //  n = repeat init n additional times for min / max / avg
	struct timeval t1,t2;
	long int spin_times[33][3];  // min, max,avg 
	unsigned long spin_reps[33][3];  // min, max,avg 
	int spin_bailed[32];
	int timer_count;
	int perf_calls[33];
#endif

/* Vars for the server core */
MODULE_EXPORT char *api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 0;
MODULE_EXPORT int supports_multiple = 0;
MODULE_EXPORT char *symbol_prefix = "CFontz633_";

/* Internal functions */
/* static void CFontz633_linewrap (int on); */
/* static void CFontz633_autoscroll (int on);  */
static void CFontz633_hidecursor ();
static void CFontz633_reboot (Driver * drvthis);
static void CFontz633_init_vbar (Driver * drvthis);
static void CFontz633_init_hbar (Driver * drvthis);
static void CFontz633_no_live_report ();
static void CFontz633_keys_activate();
static void CFontz633_set_temp_report (Driver* drvthis, char out[4]);
static void CFontz633_set_fan (Driver* drvthis);
static int CFontz633_packet_neq (COMMAND_PACKET *packet1, COMMAND_PACKET *packet2);
static void CFontz633_await_packet (Driver * drvthis, ubyte command);
static void CFontz633_hardware_clear (Driver * drvthis);
static void CFontz633_show_packet (Driver* drvthis, COMMAND_PACKET *packet);
static void CFontz633_ds18_init (Driver* drvthis);
static void CFontz633_fan_init (Driver* drvthis);

#ifdef TIME_INIT
static void CFontz633_init_dump_perf ();
static void CFontz633_init_perf (int timer_count, struct timeval* t1, struct timeval* t2, unsigned long spin_count, unsigned char call);
#endif

/*
 * Opens com port and sets baud correctly...
 */
MODULE_EXPORT int
CFontz633_init (Driver * drvthis, char *args)
{
	struct termios portset;
	int tmp, w, h;
	int reboot = 0;
	int usb = 0;

	int contrast = DEFAULT_CONTRAST;
	char device[200] = DEFAULT_DEVICE;
	int speed = DEFAULT_SPEED;
	char size[200] = DEFAULT_SIZE;
	/*
	 * drvthis633 declared extern in io.c
	 */
	drvthis633=drvthis;
	
	debug(RPT_INFO, "CFontz633: init(%p,%s)", drvthis, args );

	EmptyKeyRing();
	EmptyReceiveBuffer();

	/* Read config file */
	/* Which serial device should be used */
	strncpy(device, drvthis->config_get_string ( drvthis->name , "Device" , 0 , DEFAULT_DEVICE),sizeof(device));
	device[sizeof(device)-1]=0;
	debug (RPT_INFO,"CFontz633: Using device: %s", device);

	/* Which size */
	strncpy(size, drvthis->config_get_string ( drvthis->name , "Size" , 0 , DEFAULT_SIZE),sizeof(size));
	size[sizeof(size)-1]=0;
	if( sscanf(size , "%dx%d", &w, &h ) != 2
	|| (w <= 0) || (w > LCD_MAX_WIDTH)
	|| (h <= 0) || (h > LCD_MAX_HEIGHT)) {
		report (RPT_WARNING, "CFontz633_init: Cannot read size: %s. Using default value.\n", size);
		sscanf( DEFAULT_SIZE , "%dx%d", &w, &h );
	} else {
		width = w;
		height = h;
	}

	/* Which contrast */
	if (0<=drvthis->config_get_int ( drvthis->name , "Contrast" , 0 , DEFAULT_CONTRAST) && drvthis->config_get_int ( drvthis->name , "Contrast" , 0 , DEFAULT_CONTRAST) <= 1000) {
		contrast = drvthis->config_get_int ( drvthis->name , "Contrast" , 0 , DEFAULT_CONTRAST);
	} else {
		report (RPT_WARNING, "CFontz633_init: Contrast must between 0 and 1000. Using default value.\n");
	}

	/* Which backlight brightness */
	if (0<=drvthis->config_get_int ( drvthis->name , "Brightness" , 0 , DEFAULT_BRIGHTNESS) && drvthis->config_get_int ( drvthis->name , "Brightness" , 0 , DEFAULT_BRIGHTNESS) <= 100) {
		brightness = drvthis->config_get_int ( drvthis->name , "Brightness" , 0 , DEFAULT_BRIGHTNESS);
	} else {
		report (RPT_WARNING, "CFontz633_init: Brightness must between 0 and 100. Using default value.\n");
	}

	/* Which backlight-off "brightness" */
	if (0<=drvthis->config_get_int ( drvthis->name , "OffBrightness" , 0 , DEFAULT_OFFBRIGHTNESS) && drvthis->config_get_int ( drvthis->name , "OffBrightness" , 0 , DEFAULT_OFFBRIGHTNESS) <= 100) {
		offbrightness = drvthis->config_get_int ( drvthis->name , "OffBrightness" , 0 , DEFAULT_OFFBRIGHTNESS);
	} else {
		report (RPT_WARNING, "CFontz633_init: OffBrightness must between 0 and 100. Using default value.\n");
	}


	/* Which speed */
	tmp = drvthis->config_get_int ( drvthis->name , "Speed" , 0 , DEFAULT_SPEED);
	if (tmp == 1200) speed = B1200;
	else if (tmp == 2400) speed = B2400;
	else if (tmp == 9600) speed = B9600;
	else if (tmp == 19200) speed = B19200;
	else { report (RPT_WARNING, "CFontz633_init: Speed must be 1200, 2400, 9600 or 19200, Using default value.\n", speed);
	}

	/* New firmware version?
	 * I will try to behave differently for firmware 0.6 or above.
	 * Currently this is not in use.
	 */
	if(drvthis->config_get_bool( drvthis->name , "NewFirmware" , 0 , 0)) {
		newfirmware = 1;
	}

	/* Reboot display? */
	if (drvthis->config_get_bool( drvthis->name , "Reboot" , 0 , 0)) {
//		report (RPT_INFO, "LCDd: rebooting CrystalFontz LCD...\n");
		reboot = 1;
	}

	/*Am I USB or not?*/
	if(drvthis->config_get_bool( drvthis->name , "USB" , 0 , 0)) {
		usb = 1;
	}
	
	/*  Temp and Fan syslog reporting or no */
	syslogsensors=drvthis->config_get_int(drvthis->name, "SyslogSensors" , 0 , DEFAULT_SYSLOG_SENSORS ); 

	/* Set up io port correctly, and open it... */
	debug( RPT_DEBUG, "CFontz633: Opening serial device: %s", device);
    if ( usb ) {
        fd = open (device, O_RDWR | O_NOCTTY);
    } else {
        fd = open (device, O_RDWR | O_NOCTTY | O_NDELAY);
    }
	if (fd == -1) {
		report (RPT_ERR, "CFontz633_init: failed (%s)\n", strerror (errno));
		return -1;
	}
	if (fd < 0) {
		report (RPT_ERR, "CFontz633_init: failed fd=open (%s)\n", strerror (errno));
		return -1;
	}

	tcgetattr (fd, &portset);

	/* We use RAW mode */
    if ( usb ) {
        // The USB way
        portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
                              | INLCR | IGNCR | ICRNL | IXON );
        portset.c_oflag &= ~OPOST;
        portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
        portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
        portset.c_cflag |= CS8 | CREAD | CLOCAL ;
        portset.c_cc[VMIN] = 0;
        portset.c_cc[VTIME] = 0;
    } else {
#ifdef HAVE_CFMAKERAW
		/* The easy way */
		cfmakeraw( &portset );
#else
		/* The hard way */
		portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
	                      | INLCR | IGNCR | ICRNL | IXON );
		portset.c_oflag &= ~OPOST;
		portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
		portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
		portset.c_cflag |= CS8 | CREAD | CLOCAL ;
#endif
    }

	/* Set port speed */
	cfsetospeed (&portset, speed);
	cfsetispeed (&portset, B0);

	/* Do it... */
	tcsetattr (fd, TCSANOW, &portset);

	/* Make sure the frame buffer is there... */
	framebuf = (unsigned char *) malloc (width * height);
	memset (framebuf, ' ', width * height);

	/* Set display-specific stuff.. */
#ifdef TIME_INIT
	for(tmp=0; tmp <32; tmp++){
		perf_calls[tmp]=0;
		spin_bailed[tmp]=0;
		int tmp2;
		for(tmp2=0; tmp2<4; tmp2++){
			spin_times[tmp][tmp2]=0;
			spin_reps[tmp][tmp2]=0;
		}
	}
	for (timer_count=0; timer_count< time_init; timer_count++) {
#endif
	if (reboot) {
		report (RPT_INFO, "LCDd: rebooting CrystalFontz LCD...");
		CFontz633_reboot (drvthis);
		report (RPT_INFO, "LCDd: CrystalFontz LCD rebooted");
		reboot = 0;
	} // end if (reboot)

	CFontz633_no_live_report ();

	/* Do temperatures or no */
	tempcount=drvthis->config_get_int( drvthis->name , "Temperatures" , 0 , DEFAULT_TEMP_REPORT);  

	if (tempcount >0) {
	//	report (RPT_INFO, "LCDd: CFontz633_ds18_init started...\n");
		CFontz633_ds18_init (drvthis);
	}  

	CFontz633_fan_init (drvthis);

	CFontz633_hidecursor ();

	CFontz633_set_contrast (drvthis, contrast);

	CFontz633_keys_activate();

// show a line of syslog column names -- sync with end of treat _packet
	char syslogline[1023];
	int lineused=0;
	report(RPT_DEBUG, "CFontz633.c: building logline: \n");
	for (tmp=0; tmp <4; tmp++) {
		if (CF633fans[tmp].p_log) {
			sprintf (&syslogline[lineused], "%s:pwr,", CF633fans[tmp].name);
			lineused += strlen (CF633fans[tmp].name) +5;
		}
		if (CF633fans[tmp].autoreport) {
			sprintf (&syslogline[lineused], "%s:%s,", CF633fans[tmp].name, CF633fans[tmp].d_label);
			lineused += strlen (CF633fans[tmp].name);
			lineused += strlen (CF633fans[tmp].d_label) +2;
		}
	} // end of build syslogline for each fan
	if (tempcount > 0) {
		for (tmp=0; tmp < tempcount; tmp++) {
				sprintf (&syslogline[lineused], "%s,", CF633temps[tmp].name);
				lineused += strlen (CF633temps[tmp].name) +1;
		} // end of build syslogline for each Temperature	
	} // end of build syslogline for all Temperatures
	report(RPT_CRIT, "CF633data,%s", syslogline);

#ifdef TIME_INIT // close & free everything that will be recreated
report(RPT_INFO, "CFontz633: closing TIME_INIT %d", timer_count+1);
if (timer_count!=time_init-1){ // unless this is the last recreation
	if (tempcount !=0 ) {
		unsigned char sense_disable_mask[4]={0,0,0,0};
		CFontz633_set_temp_report(drvthis, sense_disable_mask);
		if (CF633temps) {
			int counter=0;
			for (counter=0; counter < tempcount; counter ++){
				if (CF633temps[counter].name){
					free (CF633temps[counter].name);
				}
			}
			free (CF633temps);
		} // end if (CF633temps)
	} // end if tempcount !=0
	if (CF633fans) {
			int counter=0;
			for (counter=0; counter < 4; counter ++){
				if (CF633fans[counter].name) free (CF633fans[counter].name);
				if (CF633fans[counter].d_label) free (CF633fans[counter].d_label);
			}
			free (CF633fans);
	} // end if (CF633fans)
	send_onebyte_message(fd,  CF633_Set_Up_Fan_Reporting, 0);
	CFontz633_await_packet (drvthis,  CF633_Set_Up_Fan_Reporting); 
}
reboot=1;
	} // end of big timer_count iteration that started b4 if (boot)
	CFontz633_init_dump_perf ();
#endif
	init_completed = 1;  //let check for packet start storing & logging

	return 0;
} //	end of CFontz633_init	

/*
 * Clean-up
 */
MODULE_EXPORT void
CFontz633_close (Driver * drvthis)
{
		int tempy=0; // does screen_dest takes widets with it
		for (tempy=0; tempy <  CF633ScreenCount; tempy ++){
				screen_destroy(CF633Screens[tempy]);
		}
		
	if (tempcount !=0 ) {
		unsigned char sense_disable_mask[4]={0,0,0,0};
		CFontz633_set_temp_report(drvthis, sense_disable_mask);
		if (CF633temps) {
			int counter=0;
			for (counter=0; counter < tempcount; counter ++){
				if (CF633temps[counter].name){
					free (CF633temps[counter].name);
				}
			}
			free (CF633temps);
			CF633temps = NULL;
		} // end if (CF633temps)
	} // end if tempcount !=0
	if (CF633fans) {
			int counter=0;
			for (counter=0; counter < 4; counter ++){
				if (CF633fans[counter].name) free (CF633fans[counter].name);
				if (CF633fans[counter].d_label) free (CF633fans[counter].d_label);
			}
			free (CF633fans);
			CF633fans = NULL;
	} // end if (CF633fans)
	send_onebyte_message(fd,  CF633_Set_Up_Fan_Reporting, 0);
	CFontz633_await_packet ( drvthis, CF633_Set_Up_Fan_Reporting); 
	
	close (fd);

	if(framebuf) free (framebuf);
	framebuf = NULL;
	
	int tmp;
	report(RPT_INFO, "CFontz633: other_packets tally: ");
	for (tmp=0;tmp <64; tmp++) {
		if (other_packets[tmp]>0) report(RPT_INFO, "CFontz633: [%d 0x%x]%s    %d\n",
				tmp, tmp, cf633_pack_names[tmp], other_packets[tmp]);
	}

	if(old) free (old);
	old = NULL;
}

/*
 * Returns the display width
 */
MODULE_EXPORT int
CFontz633_width (Driver *drvthis)
{
/*
        PrivateData *p = (PrivateData *) drvthis->private_data;
	return p->width;
*/
	return width;
}

/*
 * Returns the display height
 */
MODULE_EXPORT int
CFontz633_height (Driver *drvthis)
{
	return height;
}

/*
 * Flushes all output to the lcd...
 */
MODULE_EXPORT void
CFontz633_flush (Driver * drvthis)
{
	int i;
	char *xp, *xq;
/*
 * We don't use delta update yet.
 * It is possible but not easy, we can only update a line, full or begining.
 */
	if (old==NULL) {
		old = (unsigned char *) malloc (width * height);
		memset (old, ' ', width * height);
		CFontz633_hardware_clear (drvthis);
	}

xp = framebuf;
xq = old;

for (i=0; i<width; i++) {
	if (*xp != *xq) {
send_bytes_message(fd, 16, CF633_Set_LCD_Contents_Line_One, &framebuf[0]);
//CFontz633_await_packet (drvthis,   CF633_Set_LCD_Contents_Line_One);
// await_packet:critical
memcpy(old, framebuf, width);
		break;
		}
	xp++; xq++;
	}

xp = &framebuf[width];
xq = &old[width];

for (i=0; i<width; i++) {
	if (*xp != *xq) {
send_bytes_message(fd, 16, CF633_Set_LCD_Contents_Line_Two, &framebuf[width]);
//CFontz633_await_packet (drvthis,   CF633_Set_LCD_Contents_Line_One);
// await_packet:critical
memcpy(&old[width], &framebuf[width], width);
		break;
		}
	xp++; xq++;
	}

}

/*
 * Return one char from the KeyRing
 */
MODULE_EXPORT char *
CFontz633_get_key (Driver *drvthis)
{
        unsigned char akey;

	akey = GetKeyFromKeyRing();

        switch(akey) {
                case CF633_KEY_LEFT:
                        return "Left";
                        break;
                case CF633_KEY_UP:
                        return "Up";
                        break;
                case CF633_KEY_DOWN:
                        return "Down";
                        break;
                case CF633_KEY_RIGHT:
                        return "Right";
                        break;
                case CF633_KEY_ENTER:
                        return "Enter"; /* Is this correct ? */
                        break;
                case CF633_KEY_ESCAPE:
                        return "Escape";
                        break;
                default:
                        if (akey !=0) report( RPT_INFO, "cfontz633: Untreated key 0x%2x", akey); 
                        return NULL;
                        break;
        }
}


/*
 * Prints a character on the lcd display, at position (x,y).
 * The upper-left is (1,1), and the lower right should be (16,2).
 */
MODULE_EXPORT void
CFontz633_chr (Driver * drvthis, int x, int y, char c)
{
	y--;
	x--;

	framebuf[(y * width) + x] = c;
}

/*
 * Returns current contrast
 * This is only the locally stored contrast, the contrast value
 * cannot be retrieved from the LCD.
 * Value 0 to 1000.
 */
MODULE_EXPORT int
CFontz633_get_contrast (Driver * drvthis)
{
	return contrast;
}

/*
 *  Changes screen contrast (valid hardware value: 0-50)
 *  Value 0 to 1000.
 */
MODULE_EXPORT void
CFontz633_set_contrast (Driver * drvthis, int promille)
{
	int hardware_contrast;
	/* Check it */
	if( promille < 0 || promille > 1000 )
		return;

	/* Store the software value since there is not get. */
	contrast = promille;

	hardware_contrast = contrast/20;
/* Next line is to be checked $$$ */
	send_onebyte_message(fd, CF633_Set_LCD_Contrast, hardware_contrast);
	CFontz633_await_packet (drvthis,   CF633_Set_LCD_Contrast); 
}

/*
 * Sets the backlight on or off.
 * The hardware support any value between 0 and 100.
 * Need to find out if we have support for intermediate value.
 */
MODULE_EXPORT void
CFontz633_backlight (Driver * drvthis, int on)
{
	if (on) {
/* Next line is to be checked $$$ */
send_onebyte_message(fd, CF633_Set_LCD_And_Keypad_Backlight, brightness);
//CFontz633_await_packet (drvthis,   CF633_Set_LCD_And_Keypad_Backlight); 
// await_packet:critical
} else {
/* Next line is to be checked $$$ */
send_onebyte_message(fd, CF633_Set_LCD_And_Keypad_Backlight, offbrightness);
//CFontz633_await_packet (drvthis,  CF633_Set_LCD_And_Keypad_Backlight); 
// await_packet:critical
	}
}

/*
 * Get rid of the blinking curson
 */
static void
CFontz633_hidecursor ()
{
	send_onebyte_message(fd, CF633_Set_LCD_Cursor_Style, 0);
//	CFontz633_await_packet (drvthis,   CF633_Set_LCD_Cursor_Style); 
}

/*
 * Stop live reporting of temperature.
 */
static void
CFontz633_no_live_report ()
{
	char out[2]= {0, 0};
	for (out[0]=0; out[0]<8; out[0]++){
	    	send_bytes_message(fd, 2, CF633_Set_Up_Live_Fan_or_Temperature_Display , out);
		//CFontz633_await_packet (drvthis,   CF633_Set_Up_Live_Fan_or_Temperature_Display); 
	}
}

/*
 * Tell 633 to report key presses.
 */
static void
CFontz633_keys_activate()
{
	// set keypress mask
	char out[2]= {63, 0};  // presses only
	//char out[2]= {0, 63};  // releases only
	//char out[2]= {63, 63};  // both 
	send_bytes_message(fd, 2, CF633_Configure_Key_Reporting, out);

	//CFontz633_await_packet (drvthis,   CF633_Configure_Key_Reporting); 
}


/*
 * Detect ds18 temperature sensors.
 */
static void 
CFontz633_ds18_init (Driver* drvthis)
{
	unsigned char retry=3;
	unsigned char this_retry=0;

	unsigned char sensors=0;
	unsigned char detected=0;
	unsigned char sense_enable_mask[4]={0,0,0,0};
	unsigned char sensor_mask_thingy=1;
	do {
		report (RPT_DEBUG, "CFontz633_init: CFontz633_ds18_init: sending[%d]", sensors);
		send_onebyte_message(fd, CF633_Read_DOW_Device_Information, sensors);
		if (sensors==0) usleep(100000);  //  more likely to detect all on 1st try with a pause ?
			// is the long pause for 1st report dependent on # of sensors?
			//  usleep(005000*tempcount) perhaps
		report (RPT_DEBUG, "CFontz633_init: CFontz633_ds18_init: sent[%d]", sensors);
		do {
			CFontz633_await_packet ( drvthis, CF633_Read_DOW_Device_Information); 
		} while (incoming_command.data[0] !=sensors );
		report (RPT_DEBUG, "CFontz633_init: CFontz633_ds18_init: wait got [%d]", sensors);
		if ((incoming_command.data[1]!=0) && (sensors < tempcount)) {
//			report (RPT_INFO, "CFontz633_init: CFontz633_ds18_init: sense_enable_mask[%d] += %3d",
//				(int)(sensors/8),     sensor_mask_thingy );
			sense_enable_mask[(int)(sensors/8)]+= sensor_mask_thingy ;
			if (sensor_mask_thingy == 128 ) sensor_mask_thingy=1; 
				else sensor_mask_thingy*=2;
			detected ++;
			sensors ++;
		} // endif do not enable unreqested sensors
		else if ( (incoming_command.data[1]!=0) && (sensors >= tempcount)) {
			detected ++;
			sensors ++;
		}
		else if ( (incoming_command.data[1]==0) && (sensors >= tempcount)) {
			sensors ++;
		}
		if ((incoming_command.data[1]==0) && (tempcount > detected ) && (this_retry < retry -1  ) ){
			report (RPT_INFO, "CFontz633_init: CFontz633_ds18_init: detected %d of suggested %d Temperatures, attempt # %d, retry."
				, detected, tempcount, this_retry );
//			report (RPT_INFO, "CFontz633_init: CFontz633_ds18_init: sense_enable_mask built to [%d %d %d %d] ",
//				sense_enable_mask[0],sense_enable_mask[1],sense_enable_mask[2],sense_enable_mask[3] );
			sensors=0;
			detected=0;
			sense_enable_mask[0]=0;
			sense_enable_mask[1]=0;
			sense_enable_mask[2]=0;
			sense_enable_mask[3]=0;
//			report (RPT_INFO, "CFontz633_init: CFontz633_ds18_init, sense_enable_mask reset to [%d %d %d %d] ",
//				sense_enable_mask[0],sense_enable_mask[1],sense_enable_mask[2],sense_enable_mask[3] );
			sensor_mask_thingy=1;
			this_retry++;
		}
		else if ( (incoming_command.data[1]==0)  && ( this_retry == retry -1 )){
			report (RPT_INFO, "CFontz633_init: CFontz633_ds18_init: detected %d of suggested %d Temperatures, attempt # %d, no retry."
				, detected, tempcount, this_retry );
			this_retry++;
		}
//		report (RPT_INFO, "CFontz633_init: this_retry %d sensors %d", this_retry, sensors);
	} while (( this_retry < retry  ) && (detected < tempcount ));
	
	if (tempcount == detected) { 
		report (RPT_INFO, "CFontz633_init: %d Temperatures requested;  %d Temperatures detected, attempt %d."
			, tempcount, detected, this_retry );
	}
	else if (tempcount > detected) { 
		report (RPT_WARNING, "CFontz633_init: %d Temperatures suggested; continuing with %d Temperatures detected."
			, tempcount, detected  );
		tempcount=detected;
	}
	else if (tempcount < detected) { 
		report (RPT_WARNING, "CFontz633_init: %d Temperatures detected; continuing with %d Temperatures enabled as requested."
			,  detected, tempcount );
	}
//	report (RPT_INFO, "CFontz633_init: sense_enable_mask[%3d][%3d][%3d][%3d]", 
//		sense_enable_mask[0], sense_enable_mask[1], sense_enable_mask[2], sense_enable_mask[3]);
	if (tempcount !=0) {
		if ((CF633temps=malloc(tempcount*sizeof(CF_633_DS18)))) {
			char buf[255]; // BUFSIZE is 255 from ../../shared/configfile.c
			char target[64];
			for (sensors =0; sensors < tempcount; sensors++){ 
				sprintf(target, "633T%d_Name", sensors+1);	
				CF633temps[sensors].aswidget=NULL;
				CF633temps[sensors].ds18= DEFAULT_DS18_TEMP; 
				snprintf(buf, sizeof(buf), drvthis->config_get_string ( drvthis->name , target, 0 , target));
				if ((strlen(buf)==0) || (*buf==*target))  // empty DEFAULT_DS18_NAME line from config
					sprintf (buf, "T%02d", sensors+1); // add sensor number to default name
				if ((CF633temps[sensors].name =(char *)malloc (strlen (buf)+1)))
					sprintf(CF633temps[sensors].name, buf);
				else report (RPT_WARNING, "CFontz633_init: sensor %d was not malloced  a name."); 
//				report (RPT_INFO, "CFontz633_init: sensor %d name %s", sensors+1, CF633temps[sensors].name);

				sprintf(target, "633T%d_lcd", sensors+1);	
				CF633temps[sensors].dowidget=drvthis->config_get_bool(drvthis->name, target , 0, DEFAULT_WIDGET);
			}
		}
		else {
			tempcount=0;
			report (RPT_WARNING, "CFontz633_init: Temperatures malloc failed, temperature reporting disabled.");
		}
	}
	if (tempcount > 0) {
		CFontz633_set_temp_report(drvthis, sense_enable_mask);
	}
} // end ds18_init

#ifdef TIME_INIT
/* 
 * take minimim, makimum, and average time of things, 
 */
static void 
CFontz633_init_perf (int timer_count, struct timeval* t1, struct timeval* t2,  unsigned long spin_count, unsigned char call ){
	long int t_diff;
	t_diff = ((t2->tv_sec - t1->tv_sec) * 1e6 + (t2->tv_usec - t1->tv_usec));
	if (t_diff<spin_times[call][0]) spin_times[call][0]=t_diff;
	if (t_diff>spin_times[call][1]) spin_times[call][1]=t_diff;
	spin_times[call][2]=((timer_count*spin_times[call][2]+t_diff)/(timer_count +1));
	if (spin_count < spin_reps[call][0]) spin_reps[call][0]=spin_count;
	if (spin_count > spin_reps[call][1]) spin_reps[call][1]=spin_count;
	spin_reps[call][2]=((timer_count*spin_reps[call][2]+spin_count)/(timer_count +1));

	if (timer_count==0) {   // if first call, overwrite zero initialized minimum
		spin_times[call][0]=t_diff;
		spin_reps[call][0]=spin_count;
	}
	perf_calls[call]++;
	
} //end of CFontz633_init_perf

/* 
 * display minimum, maximum, and average time of things, 
 */
static void 
CFontz633_init_dump_perf (){
/*	long int spin_times[32][3];  // min, max,avg 
	int spin_reps[32][3];  // min, max,avg  */
	int tmp;

	report (RPT_CRIT, "calls	min s		max s 		avg s		min		max	 	 avg spin 	decription.");
	for (tmp =0; tmp<32; tmp++) {
		if ((spin_times[tmp ][0]!=0)||(spin_times[tmp ][1]!=0)||(spin_times[tmp ][2]!=0)){
			report (RPT_CRIT, "%d	%2.6f	%2.6f	%2.6f	%8ld	%8ld	%8ld	%s.",
				 perf_calls[tmp] , ((float)spin_times[tmp ][0]/1000000)   
				, ((float)spin_times[tmp ][1]/1000000)   
				, ((float)spin_times[tmp ][2]/1000000) 
				, spin_reps[tmp ][0], spin_reps[tmp ][1], spin_reps[tmp ][2], 
				cf633_pack_names[tmp]);
		}
	}
} // end CFontz633_init_dump_perf
#endif
/* 
 * parse fan configs, verify, create *CF633fans
 */
static void 
CFontz633_fan_init (Driver* drvthis){
	char target[64]; 
	int fans=0;
	int tmp;
	char buf[255]; // BUFSIZE is 255 from ../../shared/configfile.c
	if ((CF633fans=malloc(4*sizeof(CF_633_FAN)))) {
  		for (fans =0; fans <  4; fans++){ 
			CF633fans[fans].name=NULL;
			CF633fans[fans].aswidget[0]=NULL;
			CF633fans[fans].aswidget[1]=NULL;
			char fan_safety_checks=1;
			sprintf (target, "633F%d_Name", fans+1);	
			sprintf(buf, drvthis->config_get_string ( drvthis->name , target, 0 , DEFAULT_FAN_NAME));
			if (strlen(buf)==0)  // empty 633F_Name line in config
					sprintf(buf, DEFAULT_FAN_NAME);
			if (!strcmp(buf,DEFAULT_FAN_NAME))
				sprintf (buf+ strlen(DEFAULT_FAN_NAME), "%d", fans+1); // add fan number to default name
			if ((CF633fans[fans].name =(char *)malloc (strlen (buf)+1)))
				sprintf(CF633fans[fans].name, buf);
			else report (RPT_WARNING, "CFontz633_init: 633F%d was not malloced  a name.", fans+1); 
			if (strncmp(buf,"^",1)==0) {
				report (RPT_INFO, "CFontz633_init: 633F%d safety checking disabled.", fans+1); 
				fan_safety_checks=0;
			}

			sprintf (target, "633F%d_D_label", fans+1);	
			sprintf(buf, drvthis->config_get_string ( drvthis->name , target, 0 , DEFAULT_FAN_D_LABEL));
			if (strlen(buf)==0)  // empty 633F_Name line in config
				sprintf(buf, DEFAULT_FAN_D_LABEL);
			if ((CF633fans[fans].d_label =(char *)malloc (strlen (buf)+1)))
				sprintf(CF633fans[fans].d_label, buf);
			else report (RPT_WARNING, "CFontz633_init: 633F%d was not malloced an RPM label.", fans+1); 
													
			sprintf (target, "633F%d_Rpt", fans+1);	
			CF633fans[fans].autoreport=drvthis->config_get_bool(drvthis->name, target , 0, DEFAULT_FAN_AUTOREPORT);

			sprintf (target, "633F%d_Power", fans+1);	
			tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_POWER);
			if ((tmp < 0) || (tmp > 100)) {
				report (RPT_WARNING, "CFontz633_init: 633F%d_Power=%d invalid, set to 100.", fans+1, tmp); 
				tmp=100;
			}
			CF633fans[fans].power=tmp;

			sprintf (target, "633F%d_P_Log", fans+1);	
			CF633fans[fans].p_log=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_LOG_POW);

			sprintf (target, "633F%d_Divisor", fans+1);	
			tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_DIVISOR);
			if (tmp == 0)  {
				report (RPT_WARNING, "CFontz633_init: 633F%d_Divisor=%d invalid, set to %d.",
						fans+1, tmp, DEFAULT_FAN_DIVISOR); 
				tmp=DEFAULT_FAN_DIVISOR;
			}
			CF633fans[fans].divisor=tmp;

			sprintf (target, "633F%d_Failsafe", fans+1);	
			CF633fans[fans].failsafe=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_FAILSAFE);
		
			sprintf (target, "633F%d_Glitch", fans+1);	
			CF633fans[fans].glitch=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_GLITCH);
			
			sprintf (target, "633F%d_Temp", fans+1);	
			CF633fans[fans].sensor=drvthis->config_get_int(drvthis->name, target , 0, DEFAULT_FAN_SENSOR);
			if ((CF633fans[fans].sensor  > tempcount ) || ( CF633fans[fans].sensor  < 0 )) {
				report (RPT_WARNING, "CFontz633_init: 633F%d_Temp=%d out of range: feature disabled."
						, fans+1, CF633fans[fans].sensor); 
				CF633fans[fans].sensor=DEFAULT_FAN_SENSOR;
			}
			sprintf (target, "633F%d_Ref_Temp", fans+1);	
			CF633fans[fans].sen_ref=drvthis->config_get_int(drvthis->name, target , 0, DEFAULT_FAN_SENSOR);
			if ((CF633fans[fans].sen_ref  > tempcount ) || ( CF633fans[fans].sen_ref  < 0 )) {
				report (RPT_WARNING, "CFontz633_init: 633F%d_Ref_Temp=%d out of range: feature disabled."
						, fans+1, CF633fans[fans].sen_ref); 
				CF633fans[fans].sen_ref=DEFAULT_FAN_SENSOR;
			}
			if (( CF633fans[fans].sensor!=DEFAULT_FAN_SENSOR)&&(fan_safety_checks)) { 
				if  (CF633fans[fans].p_log==0){
					CF633fans[fans].p_log=1;
					report (RPT_INFO, "CFontz633_init: 633F%d_Temp=%d causes 633F%d_Plog=yes for safety.",
							fans+1,  CF633fans[fans].sensor, fans+1); 
				}
				if  (CF633fans[fans].autoreport==0){
					CF633fans[fans].autoreport=1;
					report (RPT_INFO, "CFontz633_init: 633F%d_Temp=%d causes 633F%d_Rpt=yes for safety.",
							fans+1, CF633fans[fans].sensor, fans+1); 
				}
				if  (CF633fans[fans].failsafe==0){ 
					CF633fans[fans].failsafe=1;
					report (RPT_INFO, "CFontz633_init: 633F%d_Temp=%d causes 633F%d_Failsafe=yes for safety.",
							fans+1, CF633fans[fans].sensor, fans+1); 
				}

			} // end of simple checks for turning on sensor driven fan speed	
		
			sprintf (target, "633F%d_Abstractounits", fans+1);	
			tmp=drvthis->config_get_bool(drvthis->name, target, 0, 0);
			CF633fans[fans].t_conv=(tmp) ? 1: 16;
		
			sprintf (target, "633F%d_T_off", fans+1);	
			tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_BY_S_OFF);
			if ((tmp/CF633fans[fans].t_conv > DEFAULT_FAN_BY_S_100+DEFAULT_FAN_SAFETY) && (fan_safety_checks)) {
				report (RPT_INFO, "CFontz633_init: 633F%d_T_off=%d -> %d for safety.",
						fans+1, tmp/CF633fans[fans].t_conv, DEFAULT_FAN_BY_S_OFF); 
				CF633fans[fans].by_s_off=DEFAULT_FAN_BY_S_OFF*CF633fans[fans].t_conv;
			}
			else CF633fans[fans].by_s_off=tmp*CF633fans[fans].t_conv;

			sprintf (target, "633F%d_T_100", fans+1);	
			tmp=drvthis->config_get_int(drvthis->name, target , 0, DEFAULT_FAN_BY_S_100);
			if ((tmp/CF633fans[fans].t_conv > DEFAULT_FAN_BY_S_100+DEFAULT_FAN_SAFETY)&&(fan_safety_checks)) {
				report (RPT_INFO, "CFontz633_init: 633F%d_T_100=%d -> %d for safety.",
						fans+1, tmp/CF633fans[fans].t_conv, DEFAULT_FAN_BY_S_100); 
						CF633fans[fans].by_s_100 =DEFAULT_FAN_BY_S_100*CF633fans[fans].t_conv;
			}
			else CF633fans[fans].by_s_100=tmp*CF633fans[fans].t_conv;
			if ((CF633fans[fans].by_s_100/CF633fans[fans].t_conv <= CF633fans[fans].by_s_off/CF633fans[fans].t_conv)
					&&(fan_safety_checks)) {
				report (RPT_INFO, "CFontz633_init: 633F%d_T_100=%d is amusingly less than 633F%d_T_off=%d.",
					fans+1, CF633fans[fans].by_s_100/CF633fans[fans].t_conv, fans+1, 
					CF633fans[fans].by_s_off/CF633fans[fans].t_conv); 
			}
			sprintf (target, "633F%d_A_Sec", fans+1);	
			CF633fans[fans].a_sec=drvthis->config_get_int(drvthis->name, target , 0, DEFAULT_FAN_A_SEC);
			if ((CF633fans[fans].a_sec > 31) || (CF633fans[fans].a_sec < 1)) {
				report (RPT_INFO, "CFontz633_init: 633F%d_A_Sec=%d  not 1-31, reset to default %d.",
						fans+1, CF633fans[fans].a_sec, DEFAULT_FAN_A_SEC); 
				CF633fans[fans].a_sec=DEFAULT_FAN_A_SEC;
			}
			sprintf(target, "633F%d_lcd", fans+1);	
			CF633fans[fans].dowidget=drvthis->config_get_bool(drvthis->name, target , 0, DEFAULT_WIDGET);
			
			report (RPT_INFO, "CFontz633_init: 633F%d, %s, u %s, rpt %d, pow %d, p_log %d, div %d, fs %d, gch %d, T %s, TA %s, off %d, full %d,(%s) a_sec %d.", 
						fans+1, CF633fans[fans].name,  CF633fans[fans].d_label, 
						CF633fans[fans].autoreport, CF633fans[fans].power, CF633fans[fans].p_log,
						CF633fans[fans].divisor, CF633fans[fans].failsafe, CF633fans[fans].glitch,
						(CF633fans[fans].sensor!=0) ? CF633temps[CF633fans[fans].sensor-1].name: NULL,
						(CF633fans[fans].sen_ref!=0) ? CF633temps[CF633fans[fans].sen_ref-1].name : NULL,
						CF633fans[fans].by_s_off/CF633fans[fans].t_conv, 
						CF633fans[fans].by_s_100/CF633fans[fans].t_conv,
						(CF633fans[fans].t_conv==16) ? "deg C." : "ds18 units",
						CF633fans[fans].a_sec  )  ;	 
		} // end for each fan
	// with 633fans built, call all the send _*byte /  CF633_Set_Fan_ / await combos
	CFontz633_set_fan(drvthis);
	} // end good CF633fans=malloc
	else {
		report (RPT_WARNING, "CFontz633_init: Fan malloc failed, fan monitoritg and management disabled.");
	}	
} //end CFontz633_fan_init

/*
 * Set fan options to values in *CF633fans.
 */
static  void  
CFontz633_set_fan ( Driver* drvthis) // does it need to take an arg or just set all fans to to current CF633fans? 
{
	unsigned char fans;
	unsigned char out[4]; // need not use full length each send
	unsigned char bitmask[4];
	
	// set failsafe first seems safe
	out[0]=0;	 
	out[1]=255; // 31 7/8 seconds failsafe reset
			// could easily macro ar turn into .config option
	bitmask[0]=1;
	for (fans =0; fans <4 ;fans ++){ 
		if (CF633fans[fans].failsafe==1) out[0]+=bitmask[0];
		bitmask[0]=bitmask[0]*2;	
	}	
//	report( RPT_INFO, "CFontz633_init: CFontz633_set_fan failsafe bitmask=%d, %d.", out[0], out[1] );
	send_bytes_message(fd, 2, CF633_Set_Fan_Power_Failsafe , out);
	CFontz633_await_packet (drvthis, CF633_Set_Fan_Power_Failsafe);

	// glitch filtering 
	bitmask[0]=1;
	out[0]=0;	 
	for (fans =0; fans <4 ;fans ++){ 
		if (CF633fans[fans].glitch==1) out[0]+=bitmask[0];
		bitmask[0]=bitmask[0]*2;	
	}	
//	report( RPT_INFO, "CFontz633_init: CFontz633_set_fan glitch bitmask=%d.", out[0] );
	send_onebyte_message(fd, CF633_Set_Fan_Tachometer_Glitch_Filter , out[0]);
	CFontz633_await_packet (drvthis, CF633_Set_Fan_Tachometer_Glitch_Filter);
	
	// set fan power next CF633_Set_Fan_Power
	for (fans =0; fans <4 ;fans ++){ 
		out[fans]=CF633fans[fans].power;
	}	
//	report( RPT_INFO, "CFontz633_init: CFontz633_set_fan fan power mask=%d %d %d %d.", out[0], out[1], out[2], out[3]);
	send_bytes_message(fd, 4, CF633_Set_Fan_Power , out);
	CFontz633_await_packet ( drvthis, CF633_Set_Fan_Power);

	// set reporting
	bitmask[0]=1;
	out[0]=0;	 
	for (fans =0; fans <4 ;fans ++){ 
		if (CF633fans[fans].autoreport==1) out[0]+=bitmask[0];
		bitmask[0]=bitmask[0]*2;	
	}	
//	report( RPT_INFO, "CFontz633_init: CFontz633_set_fan autoreport bitmask=%d.", out[0] );
	send_onebyte_message(fd,  CF633_Set_Up_Fan_Reporting, out[0]);
	CFontz633_await_packet ( drvthis, CF633_Set_Up_Fan_Reporting);
} // end of CFontz633_set_fan

/*
 * Set the reporting of temperatures.
 */
static void 
CFontz633_set_temp_report (Driver * drvthis, char out[4])
{
	send_bytes_message(fd, 4, CF633_Set_Up_Temperature_Reporting, out);
	CFontz633_await_packet ( drvthis, CF633_Set_Up_Temperature_Reporting);
}   //needs an await packet ?

/*
 * Reset the display bios
 */
static void
CFontz633_reboot (Driver * drvthis)
{
	char out[3]= {8, 18, 99};
	send_bytes_message(fd, 3, CF633_Reboot, out);
	sleep(2);
	CFontz633_await_packet ( drvthis, CF633_Reboot);
//	send_zerobyte_message(fd, CF633_Ping_Command);
//	CFontz633_await_packet ( drvthis, CF633_Ping_Command);
}

/*
 * Sets up for vertical bars.
 */
static void
CFontz633_init_vbar (Driver * drvthis)
{
	char a[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
	};
	char b[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};
	char c[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};
	char d[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};
	char e[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};
	char f[] = {
		0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};
	char g[] = {
		0, 0, 0, 0, 0, 0,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
		0, 1, 1, 1, 1, 1,
	};

	if (custom != vbar) {
	//	printf("+++ vbar +++\n");
		CFontz633_set_char (drvthis, 1, a);
		CFontz633_set_char (drvthis, 2, b);
		CFontz633_set_char (drvthis, 3, c);
		CFontz633_set_char (drvthis, 4, d);
		CFontz633_set_char (drvthis, 5, e);
		CFontz633_set_char (drvthis, 6, f);
		CFontz633_set_char (drvthis, 7, g);
		custom = vbar;
	}
}

/*
 * Inits horizontal bars...
 */
static void
CFontz633_init_hbar (Driver * drvthis)
{

	char a[] = {
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
		1, 0, 0, 0, 0, 0,
	};
	char b[] = {
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
		1, 1, 0, 0, 0, 0,
	};
	char c[] = {
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
		1, 1, 1, 0, 0, 0,
	};
	char d[] = {
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
		1, 1, 1, 1, 0, 0,
	};
	char e[] = {
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
		1, 1, 1, 1, 1, 0,
	};
	char f[] = {
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1,
	};

	if (custom != hbar) {
		CFontz633_set_char (drvthis, 1, a);
		CFontz633_set_char (drvthis, 2, b);
		CFontz633_set_char (drvthis, 3, c);
		CFontz633_set_char (drvthis, 4, d);
		CFontz633_set_char (drvthis, 5, e);
		CFontz633_set_char (drvthis, 6, f);
		custom = hbar;
	}
}


/*
 * Draws a vertical bar...
 */
MODULE_EXPORT void
CFontz633_vbar (Driver * drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'up' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */

	CFontz633_init_vbar(drvthis);

	lib_vbar_static(drvthis, x, y, len, promille, options, cellheight, 0);
}


/*
 * Draws a horizontal bar to the right.
 */
MODULE_EXPORT void
CFontz633_hbar (Driver * drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'right' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */

	CFontz633_init_hbar(drvthis);

	lib_hbar_static(drvthis, x, y, len, promille, options, cellwidth, 0);
}


/*
 * Writes a big number.
 * This is not supported on 633 because we only have 2 lines...
 */
MODULE_EXPORT void
CFontz633_num (Driver * drvthis, int x, int num)
{
/*
	char out[5];
	snprintf (out, sizeof(out), "%c%c%c", 28, x, num);
	write (fd, out, 3);
*/
}

/*
 * Sets a custom character from 0-7...
 *
 * For input, values > 0 mean "on" and values <= 0 are "off".
 *
 * The input is just an array of characters...
 */
MODULE_EXPORT void
CFontz633_set_char (Driver * drvthis, int n, char *dat)
{
	char out[9];
	int row, col;
	int letter;

	if (n < 0 || n > 7)
		return;
	if (!dat)
		return;

	out[0]=n;	/* Custom char to define. xxx */

	for (row = 0; row < cellheight; row++) {
		letter = 0;
		for (col = 0; col < cellwidth; col++) {
			letter <<= 1;
			letter |= (dat[(row * cellwidth) + col] > 0);
		/* I should remove that debug code. */
		//	if (dat[(row * cellheight) + col] == 0) printf(".");
		//	if (dat[(row * cellheight) + col] == 1) printf("+");
		//	if (dat[(row * cellheight) + col] == 2) printf("x");
		//	if (dat[(row * cellheight) + col] == 3) printf("*");
		//	printf("'%1d'", dat[(row * cellwidth) + col]);
		//	printf("%3d ", letter);
		}
		out[row+1]=letter;
	//	printf(": %d\n", letter);
	}
send_bytes_message(fd, 9, CF633_Set_LCD_Special_Character_Data , out);
//CFontz633_await_packet ( drvthis, CF633_Set_LCD_Special_Character_Data);
// await_packet:critical
}

/*
 * Places an icon on screen
 */
MODULE_EXPORT int
CFontz633_icon (Driver * drvthis, int x, int y, int icon)
{
	char icons[8][6 * 8] = {
	/* Empty Heart */
		{
		 1, 1, 1, 1, 1, 1,
		 1, 1, 0, 1, 0, 1,
		 1, 0, 0, 0, 0, 0,
		 1, 0, 0, 0, 0, 0,
		 1, 0, 0, 0, 0, 0,
		 1, 1, 0, 0, 0, 1,
		 1, 1, 1, 0, 1, 1,
		 1, 1, 1, 1, 1, 1,
		 },
	/* Filled Heart */
		{
		 1, 1, 1, 1, 1, 1,		  
		 1, 1, 0, 1, 0, 1,
		 1, 0, 1, 0, 1, 0,
		 1, 0, 1, 1, 1, 0,
		 1, 0, 1, 1, 1, 0,
		 1, 1, 0, 1, 0, 1,
		 1, 1, 1, 0, 1, 1,
		 1, 1, 1, 1, 1, 1,
		 },
	/* arrow_up */
		{
		 0, 0, 0, 1, 0, 0,
		 0, 0, 1, 1, 1, 0,
		 0, 1, 0, 1, 0, 1,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 },
	/* arrow_down */
		{
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 1, 0, 1, 0, 1,
		 0, 0, 1, 1, 1, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 },
	/* checkbox_off */
		{
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 1, 1, 1, 1, 1,
		 0, 1, 0, 0, 0, 1,
		 0, 1, 0, 0, 0, 1,
		 0, 1, 0, 0, 0, 1,
		 0, 1, 1, 1, 1, 1,
		 0, 0, 0, 0, 0, 0,
		 },
	/* checkbox_on */
		{
		 0, 0, 0, 1, 0, 0,
		 0, 0, 0, 1, 0, 0,
		 0, 1, 1, 1, 0, 1,
		 0, 1, 0, 1, 1, 0,
		 0, 1, 0, 1, 0, 1,
		 0, 1, 0, 0, 0, 1,
		 0, 1, 1, 1, 1, 1,
		 0, 0, 0, 0, 0, 0,
		 },
	/* checkbox_gray */
		{
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 1, 1, 1, 1, 1,
		 0, 1, 0, 1, 0, 1,
		 0, 1, 1, 0, 1, 1,
		 0, 1, 0, 1, 0, 1,
		 0, 1, 1, 1, 1, 1,
		 0, 0, 0, 0, 0, 0,
		 },
	 /* Ellipsis */
		{
		 0, 0, 0, 0, 0, 0,		 
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 0, 0, 0, 0, 0, 0,
		 1, 0, 1, 0, 1, 0,
		 },
	};

	/* Yes we know, this is a VERY BAD implementation :-) */
	switch( icon ) {
		case ICON_BLOCK_FILLED:
			CFontz633_chr( drvthis, x, y, 255 );
			break;
		case ICON_HEART_FILLED:
		        custom = cust;
			CFontz633_set_char( drvthis, 0, icons[1] );
			CFontz633_chr( drvthis, x, y, 0 );
			break;
		case ICON_HEART_OPEN:
		        custom = cust;
			CFontz633_set_char( drvthis, 0, icons[0] );
			CFontz633_chr( drvthis, x, y, 0 );
			break;
		case ICON_ARROW_UP:
		        custom = cust;
			CFontz633_set_char( drvthis, 1, icons[2] );
			CFontz633_chr( drvthis, x, y, 1 );
			break;
		case ICON_ARROW_DOWN:
		        custom = cust;
			CFontz633_set_char( drvthis, 2, icons[3] );
			CFontz633_chr( drvthis, x, y, 2 );
			break;
		case ICON_ARROW_LEFT:
			CFontz633_chr( drvthis, x, y, 0x7F );
			break;
		case ICON_ARROW_RIGHT:
			CFontz633_chr( drvthis, x, y, 0x7E );
			break;
		case ICON_CHECKBOX_OFF:
		        custom = cust;
			CFontz633_set_char( drvthis, 3, icons[4] );
			CFontz633_chr( drvthis, x, y, 3 );
			break;
		case ICON_CHECKBOX_ON:
		        custom = cust;
			CFontz633_set_char( drvthis, 4, icons[5] );
			CFontz633_chr( drvthis, x, y, 4 );
			break;
		case ICON_CHECKBOX_GRAY:
		        custom = cust;
			CFontz633_set_char( drvthis, 5, icons[6] );
			CFontz633_chr( drvthis, x, y, 5 );
			break;
		default:
			return -1; /* Let the core do other icons */
	}
	return 0;

}

/*
 * Clears the LCD screen
 */
MODULE_EXPORT void
CFontz633_clear (Driver * drvthis)
{
	memset (framebuf, ' ', width * height);
}

/*
 * Hardware clears the LCD screen
 */
static void
CFontz633_hardware_clear (Driver * drvthis)
{
	send_zerobyte_message(fd, CF633_Clear_LCD_Screen);
//	CFontz633_await_packet (  drvthis, CF633_Clear_LCD_Screen);
// 	this is called after init_completed==1, but does not hang critically
 	}

/*
 * Prints a string on the lcd display, at position (x,y).  The
 * upper-left is (1,1), and the lower right should be (16,2).
 */
MODULE_EXPORT void
CFontz633_string (Driver * drvthis, int x, int y, char string[])
{
	int i;

	/* Convert 1-based coords to 0-based... */
	x -= 1;
	y -= 1;

	for (i = 0; string[i]; i++) {

		/* Check for buffer overflows... */
		if ((y * width) + x + i > (width * height))
			break;
		framebuf[(y * width) + x + i] = string[i];
	}
}


/* 
 * there is a built in way to comrare two structuces, yes?
 * return 1 = packets differ
 * return -1 = total eqality
*/
static int CFontz633_packet_neq (COMMAND_PACKET *packet1, COMMAND_PACKET *packet2) {
	int counter;
/*	if (packet1->CRC.as_word != packet2->CRC.as_word) return 1;	
	if (packet1->CRC.as_bytes[0] != packet2->CRC.as_bytes[0]) return 1;	
	if (packet1->CRC.as_bytes[1] != packet2->CRC.as_bytes[1]) return 1;	
*if the client / server bit of ->command is different, CRC different
*/
	if ((0x3F&(packet1->command)) != (0x3F&(packet2->command))) return 1;	
	if (packet1->data_length != packet2->data_length) return 1;	
	for (counter=0; counter < packet1->data_length; counter++)
		if (packet1->data[counter] != packet2->data[counter]) return 1;	
 	return -1;
}

/* 
 * idle until packet of type command arrives
 */
static void CFontz633_await_packet (Driver * drvthis, ubyte command) {
	static unsigned long spin_max_ever=10000;  // start with acceptable norm
	unsigned long spin_past_what=0;
	unsigned long spin_count=0;
#ifdef TIME_INIT
	gettimeofday (&t1, NULL);
#endif
	if (init_completed==1) {
		report (RPT_DEBUG, "CFontz633: await packet awaiting %s after init_completed==1.", 
			 cf633_pack_names[0x3F&(command)]);
	}
	spin_count=0;
	COMMAND_PACKET current=incoming_command;
	check_for_packet(fd, MAX_DATA_LENGTH);
	while ( (0x3F&(incoming_command.command))!=command) {
		check_for_packet(fd, MAX_DATA_LENGTH);
		if ( (CFontz633_packet_neq(&current, &incoming_command)==1) && 
		     ( (0x3F&(incoming_command.command))!=(0x3F&(command)))  )	{
			spin_past_what++;
			current=incoming_command;
			report (RPT_DEBUG, "CFontz633: await packet rep. %ld awaiting %s recieved %s.", 
				spin_past_what, cf633_pack_names[0x3F&(command)], cf633_pack_names[0x3F&(incoming_command.command)]);
			CFontz633_show_packet(drvthis, &incoming_command);
		}
		if (spin_past_what > 5) {
			report (RPT_DEBUG, "CFontz633: await packet rep. %ld awaiting %s recieved %s, and gave up.", 
				spin_past_what, cf633_pack_names[0x3F&(command)], cf633_pack_names[0x3F&(incoming_command.command)]);
#ifdef TIME_INIT
			gettimeofday (&t2, NULL);
			spin_bailed[command]++;
			CFontz633_init_perf (timer_count, &t1, &t2, spin_count , command);
#endif
			return;
		}
		spin_count++;
	}
	if (spin_count > spin_max_ever ) {
		spin_max_ever=spin_count;
		report (RPT_DEBUG, "CFontz633_await_packet new spin_max_ever %ld awaiting %s.", spin_max_ever
				, cf633_pack_names[0x3F&(command)]);
	}
#ifdef TIME_INIT
	gettimeofday (&t2, NULL);
	CFontz633_init_perf (timer_count, &t1, &t2, spin_count , command);
#endif
	return;
}

static void CFontz633_show_packet(Driver* drvthis, COMMAND_PACKET *packet)
/*
 * 2/19/05
 * something in this was crashing when run from cold boot
 * but not from warm boot, or via /etc/init.d/LCDd start | stop | restart
 * some sprintf overrun? something downstream in report?
 * something unique to me for taking a soldering iron to the com port of
 * this Gigia-Byte Nforce2 Pro?   Working now... why?
 */
{
	char target[255];
	char packstr[255];
	int usedlength=0;
	int i, cmd, top, len;

	top=(0xC0&(packet->command)) >> 6;
	cmd=(0x3F&(packet->command));
	len=packet->data_length;

	sprintf(packstr, "Packet (%02d,%02d) %02d [", top, cmd, len);
	usedlength =strlen(packstr);
	for(i=0 ;i < packet->data_length; i++){
		sprintf(target, "%02x ", packet->data[i]);
		sprintf(packstr+usedlength , "%s ",target );
		usedlength+=strlen(target);
	}
	sprintf(packstr+usedlength, "]");
	report (RPT_DEBUG, "%s", packstr); 
	report (RPT_DEBUG, "CFontz633_Show_packet strlen is %d, data_length*3 is %d", strlen(packstr), packet->data_length *3);
}
